/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
groupmod.c - Modify a group
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <accdb.h>
#include "sysprog/shared.h"
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
  E_SUCCESS = 0,
  E_OTHER,       // other error, see errno
  E_OPEN,        // unable to open back-end module or DB
  E_NOEXIST,     // group does not exist
  E_GID_USED,    // GID already used and -o was not specified
  E_NAME_USED,   // new group name (-n) already exists
  E_UPDATE,      // db->groupmod() did not return ok
  E_CLOSE,       // db->close() did not return ok
};

struct {
  char *gname, *new_gname;
  long new_gid;

  // groupmod internal
  char *ac_after, *ac_before;
  int dup, inter;
} opt = {
  gname: NULL, new_gname: NULL, new_gid: -1,
  ac_after: NULL, ac_before: NULL, dup: 0, inter: 0,
};

static int get_options(int *, const char ***);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  struct accdb_module *db;
  struct accdb_group mmask, smask, result;
  int eax, rv = E_SUCCESS;

  memset(&mmask, 0, sizeof(struct accdb_group));
  memset(&smask, 0, sizeof(struct accdb_group));
  memset(&result, 0, sizeof(struct accdb_group));
  read_config();
  if(!(eax = get_options(&argc, (const char ***)&argv))) { return !eax; }

  // ----------------------------------------
  if((db = accdb_load(module_path)) == NULL) {
    fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
     module_path, strerror(errno));
    rv = E_OPEN;
  }

  if((eax = db->open(&db->state)) <= 0) {
    fprintf(stderr, "Could not open ACCDB back-end: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OPEN;
    goto __main__close_accdb;
  }

  // ----------------------------------------
  /*
  - smask is the Search Mask used to locate a group within the UDB.
  - result is the struct which holds the information actually contained in
    the UDB. This one is not changed, but only used for comparison, i.e.
    for the -o flag.
  - mmask is the Modification Mask for groupmod().
  */
  smask.gname = opt.gname = *argv;
  smask.gid = -1;
  if((eax = db->groupinfo(db->state, &smask, &result, 1)) < 0) {
    fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OTHER;
    goto __main__close_backend;
  }

  if(eax == 0) {
    fprintf(stderr, "Group \"%s\" does not exist\n", smask.gname);
    rv = E_NOEXIST;
    goto __main__close_backend;
  }

  // ----------------------------------------
  /* In interactive mode, we start off from the results plus command line
  arguments. */
  if(opt.inter) { memcpy(&mmask, &result, sizeof(struct accdb_group)); }
  if(opt.new_gname != NULL) { mmask.gname = opt.new_gname; }
  if(opt.new_gid != -1) { mmask.gid = opt.new_gid; }

  if(opt.inter) {
    struct itab t[] = {
      {"Group name\n[%s] > ", 's', &mmask.gname, NULL},
      {"Group ID (-1 for next free)\n[%d] > ", 'd', &mmask.gid, NULL},
      {NULL},
    };

    Sh_interactive(t);
  }

  // ----------------------------------------
  if(mmask.gid != -1 && mmask.gid != result.gid && !opt.dup) {
    // If GID has changed...
    struct accdb_group qgrp = {gid: mmask.gid, gname: NULL};
    if(db->groupinfo(db->state, &qgrp, NULL, 0) > 0) {
      fprintf(stderr, "A group with GID %ld already exists."
       " Use -o to override.\n", mmask.gid);
      rv = E_GID_USED;
      goto __main__close_backend;
    }
  }

  if(mmask.gname != NULL && result.gname != NULL &&
   strcmp(mmask.gname, result.gname) != 0) {
    // ... if name changed
    struct accdb_group qgrp = {gid: -1, gname: mmask.gname};
    if(db->groupinfo(db->state, &qgrp, NULL, 0) > 0) {
      fprintf(stderr, "A group with that name (\"%s\") already exists.\n",
       mmask.gname);
      rv = E_NAME_USED;
      goto __main__close_backend;
    }
  }

  // ----------------------------------------
  if(opt.ac_before != NULL) {
    char cmd[MAXLNLEN];
    snprintf(cmd, MAXLNLEN, opt.ac_before, mmask.gname, mmask.gid);
    system(cmd);
  }

  if((eax = db->groupmod(db->state, &smask, &mmask)) <= 0) {
    fprintf(stderr, "Error: Group updating failed: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_UPDATE;
  } else if(opt.ac_after != NULL) {
    char cmd[MAXLNLEN];
    snprintf(cmd, MAXLNLEN, opt.ac_after, mmask.gname, mmask.gid);
    system(cmd);
  }

  // ----------------------------------------
__main__close_backend:
  if((eax = db->close(db->state)) <= 0) {
    fprintf(stderr, "Warning: DB was not cleanly closed:"
     " %s (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
    if(rv == E_SUCCESS) { rv = E_CLOSE; }
  }

__main__close_accdb:
  accdb_unload(db);
  return rv;
}

//-----------------------------------------------------------------------------
static int get_options(int *argc, const char ***argv) {
  static const char *_empty_argv[] = {NULL};
  const struct poptOption options_table[] = {
    // New, uudb-groupmod options
    {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_after, '\0',
     "Program to run after group modification", "cmd"},
    {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_before, '\0',
     "Program to run before group modification", "cmd"},
    {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
     "Interactively prompt for parameters", NULL},
    {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
     "Use a different module than \"*\" (the default)", "name"},

    // Default options
    {NULL, 'g', POPT_ARG_INT, &opt.new_gid, '\0',
     "Numerical value of the group's ID (UDB module might ignore this)",
     "gid"},
    {NULL, 'n', POPT_ARG_STRING, &opt.new_gname, '\0',
     "New name of the group", NULL},
    {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
     "Allow creating a group with non-unique GID"
     " (might be disabled by back-end module)", NULL},

    POPT_AUTOHELP
    POPT_TABLEEND
  };

  poptContext ctx;
  const char **args;
  int c, argk;

  ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
  poptSetOtherOptionHelp(ctx, "[OPTIONS...] group");
  while((c = poptGetNextOpt(ctx)) >= 0);
  if(c < -1) {
    fprintf(stderr, "%s: %s: %s\n", **argv, poptBadOption(ctx, 0),
     poptStrerror(c));
    return 0;
  }

  if((args = poptGetArgs(ctx)) != NULL) {
    argk = Sh_count_args(args);
    poptDupArgv(argk, args, argc, argv);
  } else {
    *argc = 0;
    *argv = _empty_argv;
  }

  if(args == NULL || *args == NULL) {
    // gname is mandatory
    poptPrintUsage(ctx, stderr, 0);
    return 0;
  }

  poptFreeContext(ctx);
  return 1;
}

static void read_config(void) {
  const char *paths[] = {"etc", "/etc", NULL};
  struct rconfig_opt options_table[] = {
    {"AC_BEFORE", 's', &opt.ac_before, NULL},
    {"AC_AFTER", 's', &opt.ac_after, NULL},
    {NULL},
  };

  accdb_rconfig_pv(paths, "groupmod.conf", options_table, RCONFIG_ONE);
  return;
}

//==[ End of file ]============================================================
