/*=============================================================================
vitalnix/accdb - User Management Suite - Unified Account Database
useradd.c - Create a new user
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <errno.h>
#include <popt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "sysprog/shared.h"
#define MAXFNLEN 1024 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

enum {
  E_SUCCESS = 0,
  E_OTHER,     // other error, see errno
  E_OPEN,      // unable to open back-end module or DB
  E_UID_USED,  // UID already used and -o was not specified
  E_NAME_USED, // user already exists
  E_UPDATE,    // db->useradd() did not return ok
  E_CLOSE,     // db->close() did not return ok
};

static struct {
  char *lname;             // login name
  long uid;                // UID, if applicable
  char *initgrp, *suppgrp, // initial (primary) and supplemental group(s)
       *gecos,             // Gecos/comment field
       *homedir,           // Home directory
       *shell;             // Shell

  char *passwd;
  int expire, inactive;

  // useradd internal
  char *ac_after, *ac_before;
  char *skeldir;
  int creat, dup, inter, split_home, sys;
} opt = {
  lname: NULL, uid: -1, initgrp: NULL, suppgrp: NULL, gecos: NULL,
  homedir: NULL, shell: "/bin/bash", passwd: NULL, expire: -1, inactive: -1,
  ac_after: NULL, ac_before: NULL, skeldir: "/etc/skel", creat: 0, dup: 0,
  inter: 0, split_home: 0, sys: 0,
};

static struct {
  long uid_min, uid_max;

  // Defaults
  char *initgrp;        // default group
  char *homebase;       // default homebase
  int expire, inactive; // and default values for expiry and inactive fields
} conf = {
  uid_min: 100,
  uid_max: 60000,

  initgrp: "100",
  homebase: "/home",
  expire: -1,
  inactive: -1,
};

static int check_login(char *);
static void generate_homedir(void);
static int get_options(int *, const char ***);
static int interactive_lname(const char *, char, void *);
static void read_config(void);

static char *module_path = "*";

//-----------------------------------------------------------------------------
int main(int argc, char **argv) {
  struct accdb_module *db;
  struct accdb_user mmask, smask;
  int eax, rv = E_SUCCESS;

  memset(&mmask, 0, sizeof(struct accdb_user));
  memset(&smask, 0, sizeof(struct accdb_user));
  read_config();
  if(!(eax = get_options(&argc, (const char ***)&argv))) { return E_OTHER; }

  opt.lname = *argv;
  if(opt.lname != NULL && !check_login(opt.lname)) { return E_OTHER; }

  // ----------------------------------------
  if((db = accdb_load(module_path)) == NULL) {
    fprintf(stderr, "Could not load ACCDB back-end module \"%s\": %s\n",
     module_path, strerror(errno));
    return E_OPEN;
  }

  if(opt.homedir == NULL) { generate_homedir(); }

  if(opt.inter) {
    struct itab t[] = {
      {"Login name\n[%s] > ", 's', &opt.lname, interactive_lname},
      {"User ID (-1 for next free)\n[%d] > ", 'd', &opt.uid, NULL},
      {"Initial group (or GID)\n[%s] > ", 's', &opt.initgrp, NULL},
      {"Supplemental groups (and/or GIDs), seperated by comma\n[%s] > ",
       's', &opt.suppgrp, NULL},
      {"Comment field (real name, etc.)\n[%s] > ", 's', &opt.gecos, NULL},
      {"Home directory\n[%s] > ", 's', &opt.homedir, NULL},
      {"Shell\n[%s] > ", 's', &opt.shell, NULL},
      {"Skeleton dir\n[%s] > ", 's', &opt.skeldir, NULL},
      {NULL},
    };

    Sh_interactive(t);
  }

  // ----------------------------------------
  if((eax = db->open(&db->state)) <= 0) {
    fprintf(stderr, "Could not open ACCDB back-end: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OPEN;
    goto __main__close_accdb;
  }

  smask.lname = opt.lname;
  smask.uid = -1;
  smask.gid = -1;

  if((eax = db->userinfo(db->state, &smask, NULL, 0)) < 0) {
    fprintf(stderr, "Error querying the ACCDB: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_OTHER;
    goto __main__close_backend;
  }

  if(eax > 0) {
    fprintf(stderr, "User \"%s\" already exists\n", smask.lname);
    rv = E_NAME_USED;
    goto __main__close_backend;
  }

  // ----------------------------------------
  if(opt.uid != -1) { // -g is provided
    smask.lname = NULL;
    smask.uid = opt.uid;
    smask.gid = -1;
    if(!opt.dup && db->userinfo(db->state, &smask, NULL, 0) > 0) {
      // The -o flag (allow creating user with duplicate GID) was not passed
      fprintf(stderr, "User with GID %ld already exists."
       " Use -o to override.\n", smask.gid);
      rv = E_UID_USED;
      goto __main__close_backend;
    }
    mmask.uid = opt.uid;
  } else if(opt.sys) { // -r flag passed
    mmask.uid = db->sysctl(ACCDB_NEXTUID_SYS, db->state);
  }

  // ----------------------------------------
  if(opt.ac_before != NULL) {
    char cmd[MAXLNLEN];
    snprintf(cmd, MAXLNLEN, opt.ac_before, mmask.lname, mmask.uid);
    system(cmd);
  }

  if(Sh_only_digits(opt.initgrp)) {
    mmask.gid = strtoul(opt.initgrp, NULL, 0);
    mmask.group = NULL;
  } else {
    mmask.gid = -1;
    mmask.group = opt.initgrp;
  }
  mmask.gecos = opt.gecos;
  mmask.passwd = "!";

/*  if((eax = db->useradd(db->state, &mmask)) <= 0) {
    fprintf(stderr, "Warning: User addition failed: %s (eax=%d, errno=%d)\n",
     strerror(errno), eax, errno);
    rv = E_UPDATE;
    goto __main__close_backend;
  } else if(opt.ac_after != NULL) {
    char cmd[MAXLNLEN];
    snprintf(cmd, MAXLNLEN, opt.ac_after, mmask.lname, mmask.uid);
    system(cmd);
  }
*/
  // XXX create home directory
  // XXX copy skeleton directory stuff

  // ----------------------------------------
__main__close_backend:
  if((eax = db->close(db->state)) <= 0) {
    fprintf(stderr, "Warning: DB was not cleanly closed: %s"
     " (eax=%d, errno=%d)\n", strerror(errno), eax, errno);
    if(rv == E_SUCCESS) { rv = E_CLOSE; }
  }

__main__close_accdb:
  accdb_unload(db);
  return rv;
}

//-----------------------------------------------------------------------------
static int check_login(char *ln) {  
  if(opt.split_home && strlen(ln) < 3) {
    fprintf(stderr, "Loginname must be at least 3 characters long\n");
    return 0;
  } else if(strlen(ln) == 0) {
    fprintf(stderr, "You need to provide a username\n");
    return 0;
  }
  return 1;
}

static void generate_homedir(void) {
  if(opt.lname != NULL && conf.homebase != NULL) {
    char buf[MAXFNLEN];
    if(opt.split_home) {
      snprintf(buf, MAXFNLEN, "%s/%c/%c%c/%s", conf.homebase, opt.lname[0],
       opt.lname[0], opt.lname[1], opt.lname);
    } else {
      snprintf(buf, MAXFNLEN, "%s/%s", conf.homebase, opt.lname);
    }
    opt.homedir = HX_strdup(buf);
  }
  return;
}

static int get_options(int *argc, const char ***argv) {
  static const char *_empty_argv[] = {NULL};
  const struct poptOption options_table[] = {
    // New, uudb-useradd options
    {NULL, 'A', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_after, '\0',
     "Program to run after user addition", "cmd"},
    {NULL, 'B', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt.ac_before, '\0',
     "Program to run before user addition", "cmd"},
    {NULL, 'I', POPT_ARG_NONE, &opt.inter, '\0',
     "Interactively prompt for parameters", NULL},
    {NULL, 'M', POPT_ARG_STRING, &module_path, '\0',
     "Use a different module than \"*\" (the default)", "name"},
    {NULL, 'S', POPT_ARG_NONE, &opt.split_home, 's',
     "Use /home/u/us/username for default home", NULL},

    // Default options
    {NULL, 'G', POPT_ARG_STRING, &opt.suppgrp, '\0',
     "Supplementary groups", "groups"}, // UNIMPLEMENTED
    {NULL, 'c', POPT_ARG_STRING, &opt.gecos, '\0',
     "User's comment in the password file", NULL},
    {NULL, 'd', POPT_ARG_STRING, &opt.homedir, '\0',
     "Home directory of the user (overrides -S)", "dir"},
    {NULL, 'e', POPT_ARG_STRING, &opt.expire, '\0',
     "Date when the account expires", "date"},
    {NULL, 'f', POPT_ARG_INT, &opt.inactive, '\0',
     "Days until account becomes inactive", "days"},
    {NULL, 'g', POPT_ARG_STRING, &opt.initgrp, '\0',
     "Initial group of the user", "group"},
    {NULL, 'k', POPT_ARG_STRING, &opt.skeldir, '\0',
     "Skeleton directory", "dir"},
    {NULL, 'm', POPT_ARG_NONE, &opt.creat, '\0',
     "Create home directory", NULL},
    {NULL, 'o', POPT_ARG_NONE, &opt.dup, '\0',
     "Allow creating a user with non-unique UID"
     " (might be disabled by back-end module)", NULL},
    {NULL, 'p', POPT_ARG_STRING, &opt.passwd, '\0',
     "Password", NULL}, // UNIMPLEMENTED
    {NULL, 'r', POPT_ARG_NONE, &opt.sys, '\0',
     "System account (use userid < UID_MIN)", NULL},
    {NULL, 's', POPT_ARG_STRING, &opt.shell, '\0',
     "The user's shell", "prog"},
    {NULL, 'u', POPT_ARG_LONG, &opt.uid, '\0',
     "Numerical value of the user's ID", "uid"},

    POPT_AUTOHELP
    POPT_TABLEEND
  };

  poptContext ctx;
  const char **args;
  int c, argk;

  ctx = poptGetContext(**argv, *argc, *argv, options_table, 0);
  while((c = poptGetNextOpt(ctx)) >= 0);
  if(c < -1) {
    fprintf(stderr, "%s: %s\n", poptBadOption(ctx, 0), poptStrerror(c));
    return 0;
  }

  if((args = poptGetArgs(ctx)) != NULL) {
    argk = Sh_count_args(args);
    poptDupArgv(argk, args, argc, argv);
  } else {
    *argc = 0;
    *argv = _empty_argv;
  }

  if(!opt.inter && (args == NULL || *args == NULL)) {
    // loginname is mandatory
    poptPrintUsage(ctx, stderr, 0);
    return 0;
  }

  poptFreeContext(ctx);
  return 1;
}

static int interactive_lname(const char *key, char type, void *ptr) {
  if(!check_login(opt.lname)) { return 0; }
  if(opt.homedir == NULL) { generate_homedir(); }
  return 1;
}

static void read_config(void) {
  const char *paths[] = {"etc", "/etc", NULL};
  struct rconfig_opt autouid_table[] = {
    {"UID_MIN", 'l', &conf.uid_min, NULL},
    {"UID_MAX", 'l', &conf.uid_max, NULL},
    {NULL},
  };
  struct rconfig_opt options_table[] = {
    {"AC_BEFORE", 's', &opt.ac_before, NULL},
    {"AC_AFTER", 's', &opt.ac_after, NULL},
    {"GROUP", 's', &conf.initgrp, NULL},
    {"HOME", 's', &conf.homebase, NULL},
    {"SHELL", 's', &opt.shell, NULL},
    {"EXPIRE", 's', &conf.expire, NULL},
    {"INACTIVE", 's', &conf.inactive, NULL},
    {"SKEL", 's', &opt.skeldir, NULL},
    // XXX split home
    {NULL},
  };

  accdb_rconfig_pv(paths, "autouid", autouid_table, RCONFIG_ONE);
  accdb_rconfig_pv(paths, "useradd.conf", options_table, RCONFIG_ONE);
  return;
}

//==[ End of file ]============================================================
